from abc import ABCMeta, abstractmethod
from json import dumps
from sys import exc_info
from traceback import print_exc

from .utils import get_status_code_string, status_codes


class HTTPException(Exception):
    def __init__(self, code, message=None, exc_info=None, headers=[]):
        Exception.__init__(self)
        self.code = code
        self.message = message
        self.exc_info = exc_info
        self.headers = headers


class ErrorHandler(metaclass=ABCMeta):
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        try:
            return self.app(environ, start_response)
        except Exception as e:
            if not isinstance(e, HTTPException):
                print_exc()
                e = HTTPException(500, message='A server error occurred. Please contact an administrator.', exc_info=exc_info())
            body, headers = self.handle(e)
            headers.extend(e.headers)
            status = get_status_code_string(e.code)
            start_response(status, headers, e.exc_info)
            return body

    @abstractmethod
    def handle(self, e):
        pass


class JSONErrorHandler(ErrorHandler):
    def __init__(self, *args, friendly=False, **kwargs):
        ErrorHandler.__init__(self, *args, **kwargs)
        self.friendly = friendly

    def handle(self, e):
        error = {
            'code': int(e.code[:3]) if isinstance(e.code, str) else e.code,
            'error': e.code[4:] if isinstance(e.code, str) else status_codes[e.code]
        }
        if e.message is not None:
            error['message'] = e.message
        return ([dumps(error, indent=4 if self.friendly else None,
                       separators=(', ', ': ') if self.friendly else (',', ':')).encode('utf-8')], [('Content-Type', 'application/json')])

# class PlainTextErrorHandler(ErrorHandler):
#     def handle(self, e):
#         return (str(e), [('Content-Type', 'text/plain')])


class HTMLErrorHandler(ErrorHandler):
    def handle(self, e):
        status_code_bytes = get_status_code_string(e.code).encode('utf-8')
        return ([b'<html><head><title>%s</title></head><body><h1>%s</h1>%s</body></html>' % (
                status_code_bytes,
                status_code_bytes,
                b'<p>%s</p>' % e.message.encode('utf-8') if e.message else b''
                )], [('Content-Type', 'text/html')])



from json import loads, JSONDecodeError
from .error import HTTPException


class Int:
    def __init__(self, min=None, max=None):
        self.min = min
        self.max = max

    def __call__(self, value):
        if value.__class__ != int:
            return False, 'expected int, found \'%s\' of type %s' % (value, value.__class__.__name__)
        elif self.min is not None and self.min > value:
            return False, 'expected int bigger than %s, found %s' % (self.min, value)
        elif self.max is not None and self.max < value:
            return False, 'expected int smaler than %s, found %s' % (self.min, value)
        else:
            return True, ''


class Float:
    def __init__(self, min=None, max=None):
        self.min = min
        self.max = max

    def __call__(self, value):
        if value.__class__ != float:
            return False, 'expected float, found \'%s\' of type %s' % (value, value.__class__.__name__)
        elif self.min is not None and self.min > value:
            return False, 'expected float bigger than %s, found %s' % (self.min, value)
        elif self.max is not None and self.max < value:
            return False, 'expected float smaler than %s, found %s' % (self.min, value)
        else:
            return True, ''


def Boolean(value):
    return value.__class__ == bool, 'expected boolean, found \'%s\' of type %s' % (value, value.__class__.__name__)


def String(value):
    return value.__class__ == str, 'expected string, found \'%s\' of type %s' % (value, value.__class__.__name__)


class List:
    def __init__(self, filter):
        self.filter = filter

    def __call__(self, value):
        if value.__class__ == list:
            for i, e in enumerate(value):
                check, reason = self.filter(e)
                if not check:
                    return False, '%s: %s' % (i, reason)
            return True, ''
        else:
            return False, 'expected list, found \'%s\' of type %s' % (value, value.__class__.__name__)


class Options:
    def __init__(self, *options):
        self.options = options

    def __call__(self, value):
        reasons = []
        for filter in self.options:
            check, reason = filter(value)
            if check:
                return True, ''
            else:
                reasons.append(reason)
        return False, 'Value not allowed (%s).' % ' or '.join(reasons)


class Object:
    def __init__(self, entries, ignore_more=False):
        self.entries = {}
        for key in entries:
            if isinstance(entries[key], tuple):
                self.entries[key] = entries[key]
            else:
                self.entries[key] = (entries[key], False)

        self.ignore_more = ignore_more

    def __call__(self, value):
        if value.__class__ == dict:
            value_keys = list(value)
            for key in self.entries:
                filter, optional = self.entries[key]
                if key in value_keys:
                    check, reason = filter(value[key])
                    if not check:
                        return False, '%s: %s' % (key, reason)
                    value_keys.remove(key)
                elif not optional:
                    return False, 'entry with key \'%s\' required' % key
            if len(value_keys) == 0 or self.ignore_more:
                return True, ''
            else:
                return False, 'unsupported key \'%s\'' % value_keys[0]
        else:
            return False, 'expected object, found \'%s\' of type %s' % (value, value.__class__.__name__)


class FilteredJSONParser:
    def __init__(self, app, filter):
        self.app = app
        self.filter = filter
        self.raw_content = None
        self.json_content = None

    def __call__(self, environ, start_response):
        if 'CONTENT_TYPE' in environ:
            if 'json' in environ['CONTENT_TYPE'].split('/')[1].split('+'):
                self.raw_content = environ['wsgi.input'].read(
                    int(environ.get('CONTENT_LENGTH', 0)))
                try:
                    self.json_content = loads(self.raw_content)
                except JSONDecodeError:
                    raise HTTPException(422, message='Invalid JSON')
                check, reason = self.filter(self.json_content)
                if check:
                    return self.app(environ, start_response)
                else:
                    raise HTTPException(400, reason)
            else:
                raise HTTPException(
                    415, message='Only json content is allowed.')
        else:
            raise HTTPException(400, message='Body required')



status_codes = {
    100: 'Continue',
    101: 'Switching Protocols',
    102: 'Processing',
    103: 'Early Hints',
    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    207: 'Multi-Status',
    208: 'Already Reported',
    226: 'IM Used',
    300: 'Multiple Choice',
    301: 'Moved Permanently',
    302: 'Found',
    303: 'See Other',
    304: 'Not Modified',
    307: 'Temporary Redirect',
    308: 'Permanent Redirect',
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Payload Too Large',
    414: 'URI Too Long',
    415: 'Unsupported Media Type',
    416: 'Range Not Satisfiable',
    417: 'Expectation Failed',
    418: 'I\'m a teapot',
    421: 'Misdirected Request',
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',
    425: 'Too Early',
    426: 'Upgrade Required',
    428: 'Precondition Required',
    429: 'Too Many Requests',
    431: 'Request Header Fields Too Large',
    451: 'Unavailable For Legal Reasons',
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
    505: 'HTTP Version Not Supported',
    506: 'Variant Also Negotiates',
    507: 'Insufficient Storage',
    508: 'Loop Detected',
    510: 'Not Extended',
    511: 'Network Authentication Required',
}


def get_status_code_string(code):
    if code in status_codes:
        return '%s %s' % (code, status_codes[code])
    else:
        return code



from base64 import b64decode
from .error import HTTPException


class BasicAuth:
    def __init__(self, app, is_correct, realm='Access to content'):
        self.app = app
        self.is_correct = is_correct
        self.realm = realm

    def __call__(self, environ, start_response):
        if 'HTTP_AUTHORIZATION' in environ and environ['HTTP_AUTHORIZATION'].startswith('Basic '):
            base64_string = environ['HTTP_AUTHORIZATION'][6:]
            try:
                byte_string = b64decode(base64_string)
                string = byte_string.decode('utf-8')
                i = string.find(':')
                if i != -1:
                    user = string[:i]
                    passwd = string[i+1:]
                    if self.is_correct(user, passwd):
                        return self.app(environ, start_response)
                    else:
                        raise HTTPException(401, message='Wrong user or password')
                else:
                    raise HTTPException(
                        400, message='Authentication not processable')
            except ValueError:
                raise HTTPException(
                    400, message='Authentication not processable')
        else:
            headers = [('WWW-Authenticate', 'Basic realm="%s"' % self.realm)]
            raise HTTPException(401, message='Authentication required', headers=headers)



from abc import ABCMeta, abstractmethod

from wsgi_tools.error import HTTPException


class Rule(metaclass=ABCMeta):
    @abstractmethod
    def get_exception(self):
        pass

    @abstractmethod
    def check(self, environ, value):
        pass

    def clear(self):
        pass


class PathRule(Rule):
    def __init__(self):
        self.args = []

    def check(self, environ, value):
        args = []
        string = True
        path = environ['PATH_INFO']
        for i in range(len(value)):
            if string:
                if path[:len(value[i])] == value[i]:
                    path = path[len(value[i]):]
                else:
                    return False
            else:
                if i+1 == len(value):
                    content = path
                    path = ''
                else:
                    content_end = path.find(value[i+1])
                    if content_end == -1:
                        return False
                    else:
                        content = path[:content_end]
                        path = path[content_end:]
                func = value[i]
                try:
                    arg = func(content)
                except ValueError:
                    return False
                args.append(arg)
            string = not string
        if path != '':
            return False
        else:
            self.args = args
            return True

    def get_exception(self):
        return HTTPException(404, message='Path not found')


class MethodRule(Rule):
    def check(self, environ, value):
        return value == environ['REQUEST_METHOD']

    def get_exception(self):
        return HTTPException(405)


METHOD_RULE = MethodRule()


class ContentTypeRule(Rule):
    def check(self, environ, value):
        if environ.get('CONTENT_TYPE') is None:
            return value is None
        elif value is None:
            return False
        if '/' in value:
            return value == environ.get('CONTENT_TYPE', 'text/plain')
        else:
            return value in environ.get('CONTENT_TYPE', 'text/plain').split('/')[1].split('+')

    def get_exception(self):
        return HTTPException(415, message='Unsupported Content-Type')


CONTENT_TYPE_RULE = ContentTypeRule()


class Router:
    def __init__(self, rules, routes):
        self.rules = rules
        self.routes = routes

    def __call__(self, environ, start_response):
        routes = list(self.routes)
        for i, rule in enumerate(self.rules):
            routes = [route for route in routes if rule.check(
                environ, route[i])]
            if len(routes) == 0:
                raise rule.get_exception()
        return self.routes[routes[0]](environ, start_response)



from functools import cached_property
from json import dumps, loads, JSONDecodeError

from wsgi_tools.error import HTTPException

from .utils import get_status_code_string


class Request:
    def __init__(self, environ):
        self.environ = environ
        self.method = environ['REQUEST_METHOD']
        self.path = environ['PATH_INFO']
        self.protocol = environ['SERVER_PROTOCOL']
        self.query_string = environ.get('QUERY_STRING', '')
        self.content_type = environ.get('CONTENT_TYPE')
        self.content_length = int(environ.get('CONTENT_LENGTH') or 0)
        self.scheme = environ['wsgi.url_scheme']
        self.body_stream = environ['wsgi.input']
        self.server = (environ['SERVER_NAME'], environ['SERVER_PORT'])
        self.headers = {}
        for key, value in environ.items():
            if key.startswith('HTTP_'):
                self.headers[key[5:].replace('_', '-')] = value

    @cached_property
    def body_bytes(self):
        return self.body_stream.read(self.content_length)

    @cached_property
    def body_string(self):
        return self.body_bytes.decode('utf-8')


class Response:
    def __init__(self, status=200, headers={}, body=[]):
        self.status = status
        self.headers = headers
        self.body = body

    def json_body(self, json, friendly=False):
        self.headers['Content-Type'] = 'application/json'
        self.body = dumps(json, indent=4 if friendly else None,
                          separators=(', ', ': ') if friendly else (',', ':'))


def _make_body(body):
    if isinstance(body, str):
        return body.encode('utf-8')
    elif isinstance(body, bytearray):
        return bytes(body)
    else:
        return body


class FriendlyWSGI:
    def __init__(self, func):
        self.func = func

    def __call__(self, environ, start_response):
        response = self.func(Request(environ))

        if isinstance(response, Response):
            status, body, headers = response.status, response.body, response.headers
        elif len(response) == 2:
            status, body = response
            headers = {}
        elif len(response) == 3:
            status, body, headers = response

        status = get_status_code_string(status)

        if isinstance(headers, dict):
            headers = [header for header in headers.items()]

        start_response(status, headers)

        if isinstance(body, (str, bytes, bytearray)):
            body = [_make_body(body)]
        else:
            body = (_make_body(d) for d in body)
        return body



from json import loads, JSONDecodeError
from .error import HTTPException
import xml.etree.ElementTree as ET

class JSONParser:
    def __init__(self, app):
        self.app = app
        self.raw_content = None
        self.json_content = None

    def __call__(self, environ, start_response):
        if 'CONTENT_TYPE' in environ:
            if 'json' in environ['CONTENT_TYPE'].split('/')[1].split('+'):
                self.raw_content = environ['wsgi.input'].read(int(environ.get('CONTENT_LENGTH', 0)))
                try:
                    self.json_content = loads(self.raw_content)
                except JSONDecodeError:
                    raise HTTPException(422, message='Invalid JSON')
                return self.app(environ, start_response)
            else:
                raise HTTPException(415, message='Only json content is allowed.')
        else:
            raise HTTPException(400, message='Body required')

class XMLParser:
    def __init__(self, app):
        self.app = app
        self.raw_content = None
        self.root_element = None

    def __call__(self, environ, start_response):
        if 'CONTENT_TYPE' in environ:
            if 'xml' in environ['CONTENT_TYPE'].split('/')[1].split('+'):
                self.raw_content = environ['wsgi.input'].read(int(environ.get('CONTENT_LENGTH', 0)))
                try:
                    self.root_element = ET.fromstring(self.raw_content)
                except ET.ParseError:
                    raise HTTPException(422, message='Invalid XML')
                return self.app(environ, start_response)
            else:
                raise HTTPException(415, message='Only xml content is allowed.')
        else:
            raise HTTPException(400, message='Body required')
